
#include "usb910conf.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbcdc.h"
#include "usb910.h"

#ifndef ADCD
#define ADCD ADC
#endif

#define AUTO_RANGE_SUPPORT
#define BIPOLAR_RANGE_SUPPORT

#define poll    usbcdc_poll

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#ifdef ADC_EXTENSION

#define ADC_MAX_CONF	8
static uint8_t adc_conf[ADC_MAX_CONF];
/* result format 
 *
 *  int16_t: unit 0.15625 mV (10/64 mV)
 *
 *   +-+-+----------+---+
 *   |1|1|  10      | 4 |
 *   +-+-+----------+---+
 *       |          |
 *
 */

#if defined(__AVR_ATtiny861__)

/* 
 *   VCC     --  ADC3 x k (about 5/2)
 *   CURRENT --  DELTA ADC0(+) ADC1(-)
 *   AIN1    --  ADC4
 *   AIN2    --  ADC5
 *   AIN3    --  ADC6
 *   AIN4    --  ADC9
 */

#define REF_VCC		(0<<REFS0)
#define REF_V110	(2<<REFS0) // need bit_clear(ADCSRB,REFS2)
#define REF_V256	(2<<REFS0) // need bit_set(ADCSRB,REFS2)
#define CHAN_ADC0	(0x00 | (1<<ADLAR))
#define CHAN_ADC1	(0x01 | (1<<ADLAR))
#define CHAN_ADC2	(0x02 | (1<<ADLAR))
#define CHAN_ADC3	(0x03 | (1<<ADLAR))
#define CHAN_VCC	CHAN_ADC3
#define CHAN_ADC4	(0x04 | (1<<ADLAR))
#define CHAN_ADC5	(0x05 | (1<<ADLAR))
#define CHAN_ADC6	(0x06 | (1<<ADLAR))
#define CHAN_ADC9	(0x09 | (1<<ADLAR))
#define CHAN_TEMP	0x1f // need bit_set(ADCSRB,MUX5)

/* all delta -- need bit_set(ADCSRB,MUX5)
 * x20/x32 : -1
 */

#define GAIN_x1		1
#define GAIN_x8		1	// need bit_set(ADCSRB,GSEL)
#define GAIN_x32	0	// need bit_set(ADCSRB,GSEL)

#define DELTA_ADC01	(0x00  | (1<<ADLAR))
#define DELTA_ADC10	(0x02  | (1<<ADLAR))
#define CHAN_CURRENT	DELTA_ADC01

#define DELTA_ADC45	(0x0c | (1<<ADLAR))
#define DELTA_ADC54	(0x0e | (1<<ADLAR))

#define DELTA_ADC56	(0x10 | (1<<ADLAR))
#define DELTA_ADC65	(0x12 | (1<<ADLAR))

#define DELTA_ADC64	(0x14 | (1<<ADLAR))
#define DELTA_ADC46	(0x16 | (1<<ADLAR))

const static uint8_t adc_chan_se[6] = 
{
	CHAN_ADC4, CHAN_ADC5, CHAN_ADC6, CHAN_ADC9,
	CHAN_ADC0, CHAN_ADC1
};

const static uint8_t adc_chan_delta[8] = 
{
	DELTA_ADC01, DELTA_ADC45, DELTA_ADC56, DELTA_ADC64,
	DELTA_ADC10, DELTA_ADC54, DELTA_ADC65, DELTA_ADC46
};


#else
#error "ADC_EXTENSION is not supported\n"
#endif

static inline uint8_t adc_is_busy() {
	return bit_is_set(ADCSRA,ADSC);
}

#define ALWAYS_ADEN

static inline uint16_t do_adc_native(uint8_t admux) {
	ADMUX = admux;
	uint16_t r1,r2,r3,tmp;
#ifdef ALWAYS_ADEN
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r1 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r2 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r3 = ADCD;
	if (r2 > r1) {
		tmp = r1; r1 = r2; r2 =tmp;
	}
	if (r3 > r1) {
		tmp = r1; r1 = r3; r3 =tmp;
	}
	if (r3 > r2) {
		tmp = r2; r2 = r3; r3 =tmp;
	}
	return r2;
#else
	ADCSRA |= (1<<ADEN);
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	/* discard 1st result */
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	ADCSRA &= ~(1<<ADEN);
	return ADCD;
#endif
}

void init_adc() {
	/* setup ADC @ 1/64 clock */
#ifdef ALWAYS_ADEN
	ADCSRA = (6<<ADPS0) | (1<<ADEN);
#else
	ADCSRA = (6<<ADPS0);
#endif
	/* digial input disable */
	cli();
	DDRA &=  ~( (1<<PA0)|(1<<PA1)|(1<<PA5)|(1<<PA6)|(1<<PA7));
	PORTA &= ~( (1<<PA0)|(1<<PA1)|(1<<PA5)|(1<<PA6)|(1<<PA7));
	sei();
	DIDR0 = (1<<ADC0D) | (1<<ADC1D) 
		| (1<<ADC4D) |(1<<ADC5D) | (1<<ADC6D);
	PORTB &= ~( (1<<PB6));
	DDRB &=  ~( (1<<PB6));
	//DIDR1 = (1<<ADC9D);
}

#if 0
void term_adc() {
	ADCSRA = 0;
	ADCSRB = 0;
	DIDR0 = 0;
	DIDR1 = 0;
}
#endif

int8_t adc_send(uint8_t size) {
	uint8_t i;
	if (size > ADC_MAX_CONF) size = ADC_MAX_CONF;
	for (i=0; i<size; i++) {
		adc_conf[i] = ext_buf[i];
	}
	return i;
}


static uint16_t vcc_val() {
	uint16_t vcc_calib;
	uint16_t result;
	uint32_t v32;
	bit_clear(ADCSRB,MUX5);
	bit_set(ADCSRB,REFS2);

	result = do_adc_native(((0<<REFS0) | CHAN_VCC));
	// range must be (1/8 to 1/2 )
	if ((result < (128<<6)) || (result >= (512<<6))) {
		return 0x7fff;
	} else {
		v32 = 0x40000000;
		vcc_calib = v32 / result; 
	}
	v32 = do_adc_native((2<<REFS0) | CHAN_VCC);
	v32 = v32 * vcc_calib;
	result = (uint16_t)(v32 >> 16);
	return result;
}

/* measure current using only single end ADC */
static uint16_t current_val2() {
	uint16_t result;
	uint32_t v32;

	v32 = vcc_val();

	bit_clear(ADCSRB,MUX5);
	result = 0xffe0 - do_adc_native((0<<REFS0) | CHAN_ADC1);
	v32 = v32 * result;
	result = (uint16_t)(v32 >> 16);
	return result;
}

static uint16_t temp_val() {
	bit_set(ADCSRB,MUX5);
	bit_clear(ADCSRB,REFS2);
	return do_adc_native((2<<REFS0) | CHAN_TEMP);
}

uint16_t do_adc_se(uint8_t chan);
int16_t do_adc_delta(uint8_t chan);

#define SEL_RANGE_V008	(uint16_t)( 0x0200 * 0.93)
#define SEL_RANGE_V032	(uint16_t)( 0x0800 * 0.93)
#define SEL_RANGE_V110	(uint16_t)( 0x4000 * 1.1 / 2.56 * 0.93)
#define SEL_RANGE_V256	(uint16_t)( 0x4000 * 0.93)

#define BSEL_RANGE_V008	(int16_t)( 0x0200 * 0.93)
#define BSEL_RANGE_V032	(int16_t)( 0x0800 * 0.93)
#define BSEL_RANGE_V110	(int16_t)( 0x4000 * 1.1 / 2.56 * 0.93)
#define BSEL_RANGE_V256	(int16_t)( 0x4000 * 0.93)

uint16_t do_adc_se(uint8_t chan) {
	uint8_t range;
	uint8_t auto_mode = 0;
	uint8_t admux;
	uint16_t result;
	range = chan & ADC_RANGE_MASK;
	if (range == ADC_RANGE_AUTO) {
		auto_mode = 1;
		range = ADC_RANGE_V512;
	}
	chan &= ADC_CHAN_MASK;
	if (chan == ADC_CHAN_VCC) {
		return vcc_val();
	}
	if (chan == ADC_CHAN_TEMP) {
		return temp_val();
	}
	if (chan == ADC_CHAN_AIN6) {
		return current_val2();
	}
	admux = adc_chan_se[chan];
	bit_clear(ADCSRB,MUX5);
	bit_clear(ADCSRB, BIN);
#ifdef AUTO_RANGE_SUPPORT
range_down:
#endif
	if (range == ADC_RANGE_V512) {
	} else if (range == ADC_RANGE_V256) {
		admux |= (2<<REFS0);
		bit_set(ADCSRB,REFS2);
	} else {
		admux |= (2<<REFS0);
		bit_clear(ADCSRB,REFS2);
	}
	result = do_adc_native(admux);
	if (range == ADC_RANGE_V512) {
		uint32_t v32;
		v32 = result;
		v32 = v32 * vcc_val();
		result = v32 >> 16;
#ifdef AUTO_RANGE_SUPPORT
		if (auto_mode) {
			if (result < SEL_RANGE_V256) {
				range = ADC_RANGE_V256;
				goto range_down;
			}
		}
#endif
	} else if (range == ADC_RANGE_V256) {
		result = result >> 2;
	} else {
#if 0
		uint32_t v32;
		v32 = (uint32_t)result;
		v32 = v32 * (uint16_t)((1.1/2.56)*0x4000);
		result = v32 >> (2 + 14);
#else
		result >>= 6; // un LAR		
		result *= 55; // 110/256 * 64 >> 2 = 55/2 >> 2
		result >>= 3;
#endif
	}
	return result;
}

int16_t do_adc_delta(uint8_t chan) {
	int16_t result;
	uint8_t range;
	int32_t v32;
	uint8_t auto_mode = 0;
	uint8_t admux;
	range = chan & ADC_RANGE_MASK;
	if (range == ADC_RANGE_AUTO) {
		auto_mode = 1;
#ifdef BIPOLAR_RANGE_SUPPORT
		if (chan & ADC_BIPOLAR) {
			range = ADC_RANGE_V256;
		} else {
			range = ADC_RANGE_V512;
		}
#else
		range = ADC_RANGE_V512;
#endif
	}
#ifdef AUTO_RANGE_SUPPORT
range_down:
#endif
	admux = adc_chan_delta[(chan & ADC_CHAN_MASK) - ADC_CHAN_DELTA];
#ifdef BIPOLAR_RANGE_SUPPORT
	if (chan & ADC_BIPOLAR) {
		bit_set(ADCSRB, BIN);
	} else {
		bit_clear(ADCSRB, BIN);
	}
#else
	bit_clear(ADCSRB, BIN);
#endif
	bit_set(ADCSRB,MUX5);
	if (range == ADC_RANGE_V512) {
		admux += GAIN_x1;
		bit_clear(ADCSRB,GSEL);
	} else if (range == ADC_RANGE_V256) {
		admux += GAIN_x1;
		admux |= (2<<REFS0);
		bit_clear(ADCSRB,GSEL);
		bit_set(ADCSRB,REFS2);
	} else if (range == ADC_RANGE_V110) {
		admux += GAIN_x1;
		admux |= (2<<REFS0);
		bit_clear(ADCSRB,GSEL);
		bit_clear(ADCSRB,REFS2);
	} else if (range <= ADC_RANGE_V032) {
		admux += GAIN_x8;
		admux |= (2<<REFS0);
		bit_set(ADCSRB,GSEL);
		bit_set(ADCSRB,REFS2);
	} else {
		admux += GAIN_x32;
		admux |= (2<<REFS0);
		bit_set(ADCSRB,GSEL);
		bit_set(ADCSRB,REFS2);
	}
	result = do_adc_native(admux);
#ifdef BIPOLAR_RANGE_SUPPORT
	if (chan & ADC_BIPOLAR) {
		if (range == ADC_RANGE_V512) {
			v32 = (result>>1);
			v32 = v32 * vcc_val();
			result = v32 >> 16;
			if ((result & 0xc000) == 0x8000) {
				result = 0x8000;
			} else if ((result & 0xc000) == 0x4000) {
				result = 0x7fff;
			} else {
				result <<= 1;
			}
		} else if (range == ADC_RANGE_V256) {
			result >>=1;
#ifdef AUTO_RANGE_SUPPORT
			if (auto_mode) {
				if ((result < BSEL_RANGE_V032) && 
					(result > -BSEL_RANGE_V032)) {
					range = ADC_RANGE_V032;
					goto range_down;
				}
			}
#endif
		} else if (range == ADC_RANGE_V110) {
#if 0
			v32 = result;
			v32 = v32 * (uint16_t)((1.1/2.56)*0x8000);
			result = v32 >> (2 + 14);
#else
			result >>= 6; // un LAR		
			result >>= 1; // avoid overflow 
			result *= 55; // 110/256 * 64 >> 1 = 55/2 >> 1
			result >>= 1;
#endif
		} else if (range <= ADC_RANGE_V032) {
			result = result >> (1 + 3);
#ifdef AUTO_RANGE_SUPPORT
			if (auto_mode && (result < BSEL_RANGE_V008) && 
					(result > -BSEL_RANGE_V008)) {
				range = ADC_RANGE_V008;
				goto range_down;
			}
#endif
		} else {
			result = result >> (1 + 5);
		}
	} else {
#endif
		if (range == ADC_RANGE_V512) {
			v32 = (uint16_t)result;
			v32 = v32 * vcc_val();
			result = (uint32_t)v32 >> 16;
#ifdef AUTO_RANGE_SUPPORT
			if (auto_mode) {
				if (result < SEL_RANGE_V032) {
					range = ADC_RANGE_V032;
					goto range_down;
				} else if (result < SEL_RANGE_V256) {
					range = ADC_RANGE_V256;
					goto range_down;
				}
			}
#endif
		} else if (range == ADC_RANGE_V256) {
			result = (uint16_t)result >> 2;
		} else if (range == ADC_RANGE_V110) {
#if 0
			v32 = (uint16_t)result;
			v32 = v32 * (uint16_t)((1.1/2.56)*0x4000);
			result = v32 >> (2 + 14);
#else
			result >>= 6; // un LAR		
			result *= 55; // 110/256 * 64 >> 2 = 55/2 >> 2
			result >>= 3;
#endif
		} else if (range <= ADC_RANGE_V032) {
			result = (uint16_t)result >> (2 + 3);
#ifdef AUTO_RANGE_SUPPORT
			if (auto_mode && (result < SEL_RANGE_V008)) {
				range = ADC_RANGE_V008;
				goto range_down;
			}
#endif
		} else {
			result = (uint16_t)result >> (2 + 5);
		}
#ifdef BIPOLAR_RANGE_SUPPORT
	}
#endif
	return result;
}

int8_t adc_recv(uint8_t size) {
	uint8_t i;
	uint16_t *p = (uint16_t *)ext_buf;
	size >>= 1;

	for (i=0; i<size; i++,p++) {
#if 1
	    if (adc_conf[i] & ADC_CHAN_DELTA) {
		    *p = do_adc_delta(adc_conf[i]);
	    } else {
		    *p = do_adc_se(adc_conf[i]);
            }
#else
		    //*p = temp_val();
		    //*p = adc_conf[i];
		    //*p = vcc_val();
		    //*p = do_adc_se(ADC_CHAN_TEMP);
		    //*p = do_adc_se(ADC_CHAN_VCC | ADC_RANGE_V512);
#endif
	}
	return (i<<1);
}

#endif
