/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*  circuit 

                           ATtiny44

                   VCC     1    14  GND
                           2    13  AREF      --- 2.5V ref
                           3    12  ADC1      --- ADC      
                   RESET   4    11  PA2 TGT_SDA   --- I2C ROM SDA
                           5    10  PA3 TGT_SCL   --- I2C ROM SCL
    PWM ----    OC0B PA7   6     9  SCK/SCL
                 MOSI/SDA  7     8  MISO


                          ----+-----------+-- VCC
                              D2 ^       (M) 
                D1            +-----------+--------+
             +--|<|--+        |                    |
             |       |        C                    R2
    PWM -----+-- R1 -+--- B                        +--+--+-----> ADC
                              E                    R3 R4 C1
                              |                    |  |  |
                           ---+--------------------+--+--+-- GND


    R1:       4.7k
    R2,R3,R4: 10k
    C1:       10nF (cutoff = 4.8kHz ?)

*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <util/delay.h>

#ifndef ADCD
#define ADCD	ADC
#endif
#ifndef ADCSRA
#define ADCSRA	ADCSR
#endif

#include "i2c.h"
#include "i2cdev.h"

#include "i2cconf.h"
uint8_t ext_buf[EXT_BUFSIZE];

#define bit_set(port,bit)       { port |= (1<<(bit));}
#define bit_clear(port,bit)     { port &= ~(1<<(bit));}

#if defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
#define CHAN_ADC1	0x01	/* ADC1 PA1 */
#define TGT_PORT	PORTA
#define TGT_DDR		DDRA
#define TGT_BIT		PA7
#endif

static inline uint8_t adc_is_busy() {
	return bit_is_set(ADCSRA,ADSC);
}

static uint16_t do_adc_native(uint8_t admux) {
	uint16_t r1,r2,r3,tmp;
	ADMUX = admux;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		;
	}
	r1 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		;
	}
	r2 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		;
	}
	r3 = ADCD;
	if (r2 > r1) {
		tmp = r1; r1 = r2; r2 =tmp;
	}
	if (r3 > r1) {
		tmp = r1; r1 = r3; r3 =tmp;
	}
	if (r3 > r2) {
		tmp = r2; r2 = r3; r3 =tmp;
	}
	return r2;
}

static inline void init_adc() {
	/* setup ADC @ 1/64 clock :125 kHz@8MHz */
	ADCSRA = (6<<ADPS0) | (1<<ADEN);
}

static inline uint16_t temp_val() {
	return 0;
//	return do_adc_native((2<<REFS0) | CHAN_TEMP);
}

static inline uint16_t adc1_val(int8_t rtype) {
#if defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
	if (rtype == 0) return do_adc_native((2<<REFS0) | CHAN_ADC1);
	if (rtype == 2) return do_adc_native((1<<REFS0) | CHAN_ADC1);
	return do_adc_native((0<<REFS0) | CHAN_ADC1);
#endif
}

#define CAPT_BUF_SIZE	64
static volatile uint8_t capt_buf[CAPT_BUF_SIZE];
static volatile uint8_t capt_rptr,capt_wptr;
static volatile int16_t capt_ov_cnt;
static uint16_t capt_len;
static uint16_t capt_cnt;
static uint8_t in_capt = 0;


static inline void capt_start(uint8_t chan,uint16_t interval) {
	ADCSRA = 0;

	if (chan == 0) ADMUX = (2<<REFS0) | CHAN_ADC1;
	else if (chan == 2) ADMUX = (1<<REFS0) | CHAN_ADC1;
	else ADMUX = (0<<REFS0) | CHAN_ADC1;

	capt_cnt = 0;
	capt_ov_cnt = 0;
	capt_rptr = capt_wptr = 0;
	in_capt = 1;

	TCNT1 = 0;
	OCR1B = 0;
	if (interval < (13*4)) interval = (13*4);
	OCR1A = interval - 1;

	TCCR1A = (0<<WGM10);
	TCCR1B = (1<<WGM12) | 2; // fast pwm @ 1Mhz 
	if (interval == (((uint16_t)OCR0A+1)<<1)) { // special case : sync with pwm 
		cli();
		TCNT1 = (uint16_t)TCNT0;
		OCR1B = OCR0B + ((OCR0A+1 - OCR0B) >> 1);
		sei();
	}
	TIMSK1 = (1<<OCIE1B); // to clear OCF1B automatically

#ifdef XXX
	ADCSRB = (0<<ADLAR) | 5; // timer1 compare match B
#else
	ADCSRB = (1<<ADLAR) | 5; // timer1 compare match B
#endif
	ADCSRA = (5<<ADPS0) | (1<<ADEN) | (1<<ADIE) | (1<<ADATE) | (1<<ADIF) | (1<<ADSC);
		// 250khz adc clock :max 19.2 k sps
}

static inline void capt_stop() {
	in_capt = 0;
	init_adc(); // re init
}

void do_capt() {
	int8_t r;
	uint8_t next;
	if ((capt_cnt & 63) == 0) {
		r = i2c_start();
		if (r != 0) return;
		ext_buf[0] = (uint8_t)(capt_cnt>>8);
		ext_buf[1] = (uint8_t)capt_cnt;
		r = i2c_send(0x50, 2);
		if (r != 2) {
			i2c_stop();
			return;
		}
	}
	next = (capt_rptr + 1)&(CAPT_BUF_SIZE -1);
	ext_buf[0] = capt_buf[capt_rptr];
	capt_rptr = next;
	i2c_send_cont(1);
	capt_cnt++;
	if ((capt_cnt & 63) == 0) {
		i2c_stop();
	}
	if (capt_cnt == capt_len) {
		capt_stop();
	}
}

SIGNAL(SIG_OUTPUT_COMPARE1B) {
	// do nothing;
}

//static int16_t prev_adc;

SIGNAL(SIG_ADC) {
	uint8_t next;
#ifdef XXX
	int16_t adc;
#endif
	next = (capt_wptr + 1) & (CAPT_BUF_SIZE -1);
	if (capt_rptr == next) {
		capt_ov_cnt++;
		return;
	}
#ifdef XXX
	adc = ADCD - prev_adc;
	if (adc < -128) adc = -128;
	if (adc > 127) adc = 127;
	capt_buf[capt_wptr] = (uint8_t)adc;
	prev_adc = ADCD;
#else
	capt_buf[capt_wptr] = ADCH;
#endif
	capt_wptr = next;
}


uint8_t myaddr;

void set_osc(uint8_t target_osccal) {
	while (OSCCAL != target_osccal) {
		if (OSCCAL < target_osccal) OSCCAL++;
		else OSCCAL--;
	}
}

//static inline 
void init_pwm() {
	bit_set(TGT_DDR, TGT_BIT);

#if defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
	OCR0B = 0x1;
	OCR0A = 99; // top 10kHz@ 8MHz clock
	TCCR0A = (2<<COM0B0) | (3<<WGM00);
	TCCR0B = (1<<WGM02) | 2; // start @ 1/8 , ctc pwm mode (TOP=OCR0A)
#else
#error "the device is not supported"
#endif
}

void set_duty(uint8_t duty) {
	if (duty > 75) duty = 75;
	if (duty == 0) duty = 1;
	OCR0B = duty;
}

int main() {
	uint8_t ch;

	/* set system_clock pre scaler 1/8 => 1/1 */
	CLKPR = (1<<CLKPCE);
	CLKPR = 0;

	myaddr = 0x10;
	i2cdev_init( myaddr );
	init_pwm();
	init_adc();
	init_i2c();

	sei();

	for(;;) {
		if( i2cdev_can_getc(1) ) {
			ch = i2cdev_getc();
			if (ch == 'S') {
				i2cdev_putc('M');
				i2cdev_putc('O');
				i2cdev_putc('T');
				i2cdev_putc('O');
				i2cdev_putc('R');
#if 0
			} else if (ch == 'O' && i2cdev_can_getc(1)) {
				set_osc(i2cdev_getc());
				i2cdev_putc(OSCCAL);
			} else if (ch == 'o') {
				i2cdev_putc(OSCCAL);
#endif
			} else if (ch == 'D' && i2cdev_can_getc(1)) {
				set_duty(i2cdev_getc());
				i2cdev_putc(OCR0B);
			} else if (ch == 'd') {
				i2cdev_putc(OCR0B);
#if 0
			} else if (!in_capt && ch == 't') {
				uint16_t temp;
				temp = temp_val();
				i2cdev_putc((uint8_t)temp);
				i2cdev_putc((uint8_t)(temp>>8));
#endif
			} else if (ch == 'v' && i2cdev_can_getc(1)) {
				uint16_t adc1;
				if (in_capt) adc1 = 0;
				else adc1 = adc1_val(i2cdev_getc());
				i2cdev_putc((uint8_t)adc1);
				i2cdev_putc((uint8_t)(adc1>>8));
			} else if (ch == 'r' && i2cdev_can_getc(2)) {
				uint8_t off_hi,off_lo,i;
				int8_t r;
				off_lo = i2cdev_getc();
				off_hi = i2cdev_getc();
				r = i2c_start();
				if (r == 0)  {
					ext_buf[0] = off_hi;
					ext_buf[1] = off_lo;
					r = i2c_send(0x50, 2);
					i2c_stop();
				}
				if (r == 2) {
					r = i2c_start();
				}
				if (r == 0) {
					r = i2c_recv(0x50, 4);
					i2c_stop();
				}
				if (r > 4) r = 4;
				i2cdev_putc(r);
				for (i=0; i<r; i++) {
					i2cdev_putc(ext_buf[i]);
				}
			} else if (ch == 'w' && i2cdev_can_getc(5)) {
				uint16_t interval;
				uint8_t chan;
				chan = i2cdev_getc();
				interval = i2cdev_getc();
				interval |= (uint16_t)(i2cdev_getc()) << 8;
				capt_len = i2cdev_getc();
				capt_len |= (uint16_t)(i2cdev_getc()) << 8;
				capt_start(chan,interval);
			} else if (ch == 's') {
				i2cdev_putc((uint8_t)capt_len);
				i2cdev_putc((uint8_t)(capt_len>>8));
				i2cdev_putc((uint8_t)capt_cnt);
				i2cdev_putc((uint8_t)(capt_cnt>>8));
				i2cdev_putc((uint8_t)capt_ov_cnt);
				i2cdev_putc((uint8_t)(capt_ov_cnt>>8));
#if 1
				i2cdev_putc(capt_wptr);
				i2cdev_putc(capt_rptr);
#else
				i2cdev_putc((uint8_t)OCR1BL);
				i2cdev_putc((uint8_t)OCR1BH);
#endif
			}
			i2cdev_flush();
		}
		if (capt_rptr != capt_wptr) {
			if (in_capt) {
				do_capt();
			} else {
				capt_rptr = capt_wptr;
			}
		}
	}
}
