/*
*  C Implementation: AVR ISP Programmer 
*
* Description:
*
*  A serial programmer  somewhat compatible with "butterfly" "avr109" and
*     "avr910" modes of avrdude
*
*
* Author: Koji Suzuki (C) 2007
* ATTiny2313_bootloader Author:
*              Ernst Bachmann <e.bachmann@netandweb.de>, (C) 2006
*
* Copyright: See License.txt file that comes with this distribution
*
*/

#include "usb910conf.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbcdc.h"
#include "usb910.h"

#if TX_SIZE < (32)
#error "TX_SIZE is too small"
#endif

#define getc	usbcdc_getc
#define putc	usbcdc_putc
#define poll	usbcdc_poll

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

static void init_timeout() {
#if defined(__AVR_ATtiny861__) || defined(__AVR_ATtiny461__)
	TCCR0A = (1<<TCW0); // 16 bit mode
	TCCR0B = 5; // timer0 start @ clk/1024
#else
	TCCR1B = 5; // timer1 start @ clk/1024
#endif
}

static inline uint16_t get_time() {
#if defined(__AVR_ATtiny861__) || defined(__AVR_ATtiny461__)
	union {
		uint16_t word;
		struct {
			uint8_t lo;
			uint8_t hi;
		};
	} timer;
	timer.lo = TCNT0L;
	timer.hi = TCNT0H;
	return timer.word;
#else
	return TCNT1;
#endif
}

static int8_t to_flag;
static int16_t to_time;

static void set_timeout(uint16_t t) {
	to_time = t + get_time();
	to_flag = 1;
}

static int8_t check_timeout() {
	if (to_flag && (to_time - (int16_t)(get_time()) <= 0)) {
		to_flag = 0;
	}
	return !to_flag;
}

static void __delay(uint16_t t) {
	set_timeout(t);
	while (check_timeout()) {
		wdt_reset();
		poll();
	}
}

/* mdelay -- max 2.79 sec @12Mhz  */
static inline void mdelay(uint8_t ms) {
	__delay((uint16_t)ms * (F_CPU/1000) / 1024);
}

static uint8_t slow_mode;
static void spi_clock(void) {
	_delay_loop_1(4*2<<slow_mode);
	bit_set(TGT_PORT, TGT_SCK);
	_delay_loop_1(4*5<<slow_mode);
	bit_clear(TGT_PORT, TGT_SCK);
	_delay_loop_1(4*3<<slow_mode);
}

static uint8_t spi_write(uint8_t c) {
	uint8_t i,rx;
	i = 0x80;
	rx = 0;
	do {
		if (c & i) {
			bit_set(TGT_PORT, TGT_MOSI);
		} else {
			bit_clear(TGT_PORT, TGT_MOSI);
		}
		if (bit_is_set(TGT_PIN, TGT_MISO)) {
			rx |= i;
		}
		spi_clock();
		i >>= 1;
	} while (i);
	return rx;
}

static inline uint8_t spi_read(void) {
	return spi_write(0x00);
}


static uint8_t target_signature[3];
static uint8_t target_pagesize;
#define pgmmode	(target_pagesize)
static uint8_t extmode = 0;
//#define EXT_BUFSIZE	16 : move to usb910.h
uint8_t ext_buf[EXT_BUFSIZE];

// tiny13   : 1e,90,07   16
//
// tiny2313 : 1e,91,0a   16
// tiny26   : 1e,91,09   16
// tiny24   : 1e,91,0b   16
// tiny25   : 1e,91,08   16
// tiny261  : 1e,91,0c   16
//
// tiny44   : 1e,92,07   32
// tiny45   : 1e,92,06   32
// tiny461  : 1e,92,08   32
//
// tiny84   : 1e,93,0c   32
// tiny85   : 1e,93,0b   32
// tiny861  : 1e,93,0d   32
// 
// mega48   : 1e,92,05   32
// mega88   : 1e,93,0a   32
// mega168  : 1e,94,06   64

static inline void get_pagesize() {
	if (target_signature[2] <= 0x91) { // 1K or 2K
		target_pagesize = 16;
	} else if (target_signature[2] <= 0x93) { // 4K or 8K
		target_pagesize = 32;
	} else if (target_signature[2] <= 0x94) { // 16K
		target_pagesize = 64;
	} else {
		target_pagesize = 128;
	}
}

static const uint8_t twd_flash = 10;
static const uint8_t twd_eeprom = 10;
static const uint8_t twd_erase = 10;

static inline void init_spi() {
	TGT_DDR &=  ~((1<<TGT_MISO));
	TGT_PORT |=  ((1<<TGT_RESET));
	TGT_PORT &=  ~((1<<TGT_SCK) | (1<<TGT_MOSI) |(1<<TGT_MISO) );
	TGT_DDR |=    ((1<<TGT_SCK) | (1<<TGT_MOSI) |(1<<TGT_RESET) );
}

static inline void term_spi() {
	TGT_DDR &=   ~((1<<TGT_SCK) | (1<<TGT_MOSI) | (1<<TGT_RESET) |(1<<TGT_MISO));
	TGT_PORT &=  ~((1<<TGT_SCK) | (1<<TGT_MOSI) | (1<<TGT_RESET) |(1<<TGT_MISO));
}

union {
	uint16_t word;
	struct {
		uint8_t lo;
		uint8_t hi;
	};
} addr;

static uint8_t do_avrisp4(uint8_t cmd,uint8_t d1,uint8_t d2,uint8_t d3) {
	spi_write(cmd);
	spi_write(d1);
	spi_write(d2);
	return spi_write(d3);
}

static uint8_t do_avrisp2(uint8_t cmd,uint8_t data) {
	return do_avrisp4(cmd, addr.hi, addr.lo, data);
}

static void enter_programming_mode() {
	int i;
	if (pgmmode != 0) return;

#ifdef OSC_EXTENSION
	term_osc();
#endif
	init_timeout();
        init_spi();

	slow_mode = 0;

	for (i=0; i<4; i++) {
		bit_clear(TGT_PORT, TGT_RESET);
		mdelay(5);
		bit_set(TGT_PORT, TGT_RESET);
		mdelay(5);
		bit_clear(TGT_PORT, TGT_RESET);
		mdelay(5);

		/* enable programming */
		spi_write(0xAC);
		spi_write(0x53);
		if (spi_write(0x00) != 0x53) {
			spi_write(0x00);
			continue;
		}
		spi_write(0x00);
		for (i =0 ; i < 3; i++) {
			target_signature[i] = do_avrisp4(0x30,0x00,i,0x00);
		}
		if (target_signature[0] == 0x1e) {
			get_pagesize();
			break;
		}
		slow_mode ++;
	}
	mdelay(5);
}

static inline void leave_programming_mode() {
        term_spi();
	pgmmode = 0;
}

int main(void)
{
    uint8_t ch,d1,d2,d3;
#ifdef AVR109_EXTENSION
    uint16_t size;
#endif


#ifndef NO_CLKPR_SET
    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = 0;
#endif

    wdt_enable(WDTO_1S);
    usbcdc_init();

    sei();

    for(;;) {
	wdt_reset();
	poll();

	ch = getc();

	if (ch == 'S') { // get Software
		putc('U');
		putc('S');
		putc('B');
		putc('9');
		putc('1');
		putc('0');
#if defined(__AVR_ATtiny861__)
		putc('A');
#else
		putc('B');
#endif
	} else if (ch == 'p') { // get Programmer Type
		putc('S'); // Serial
	} else if (ch == 'V' || ch == 'v') { // get Version 'V':sw 'v':hw
		extmode = 0;
		pgmmode = 0;
		putc('2');
		putc('4');
	} else if (ch == 'a') { // Report autoincrement address
		extmode = 0;
		pgmmode = 0;
		putc('Y'); // Yes!
	} else if (ch == 'A') { // Set Write Address
		addr.hi=getc();
		addr.lo=getc();
		putc('\r'); // ACK
	} else if (ch == 't') { // get supported Device Types
		putc(0x20); // dummy 
		putc(0);
	} else if (ch == 'T') { // set Device Types: ignore but ACK
		getc();
		putc('\r');
#if 0
	} else if (ch == 'x' || ch == 'y') { // Set/Clear LED
		putc('\r'); // ACK
#endif
	} else if (ch == 'P') { // Enter Programming mode
		enter_programming_mode();
		putc(pgmmode?'\r':'?');
	} else if (ch == 'L') { // Leave Programming mode, NOP but ACK
		leave_programming_mode();
		putc('\r');
#ifdef AVR109_EXTENSION
	} else if (ch == 'b') {// Return BufferSize
		enter_programming_mode(); // need target_pagesize
		if (pgmmode) {
			putc('Y');
			putc((target_pagesize & 0x80)?1:0);
			putc(target_pagesize<<1);
		} else {
			putc('?');
		}
#endif
	} else if ((ch == 0x1b) || (ch == '\n')) { // ESC,\n: nop
	} else if (pgmmode) {
	    if (ch == 's') { // Return Signature Byte
		putc(target_signature[2]); // part_number
		putc(target_signature[1]); // part_family
		putc(target_signature[0]); // vendor code
	    } else if (ch == 'e') { // Chip Erase
		do_avrisp4(0xAC,0x80,0x00,0x00);
		mdelay(twd_erase);
		putc('\r'); // ACK
#ifndef OFF_LEGACY
	    } else if (ch == 'c') { // Write program memory, low byte
		do_avrisp2(0x40,getc()); 
		putc('\r'); // ACK
	    } else if (ch == 'C') { // Write program memory, high byte
		do_avrisp2(0x48,getc()); 
		addr.word++;
		putc('\r'); // ACK
	    } else if (ch == 'm') { // Issue Page Write
		do_avrisp2(0x4C,0x00); 
		mdelay(twd_flash);
		putc('\r'); // ACK
	    } else if (ch == 'R') { // Read program memory
		putc(do_avrisp2(0x28,0x00)); // Read high byte
		putc(do_avrisp2(0x20,0x00)); // Read low byte
		addr.word++;
	    } else if (ch == 'D') { // Write data memory
		do_avrisp2(0xC0,getc());
		mdelay(twd_eeprom);
		addr.word++;
		putc('\r'); // ACK
	    } else if (ch == 'd') { // Read data memory
		putc(do_avrisp2(0xA0,0x00));
		addr.word++;
#endif
	    } else if (ch == 'l') { // Write lock bits
		do_avrisp4(0xAC,0xE0,0x00,getc());
		mdelay(twd_flash);
		putc('\r'); // ACK
	    } else if (ch == '.') { // New universal command
		d1 = getc();
		d2 = getc();
		d3 = getc();
		putc(do_avrisp4(d1,d2,d3,getc()));
		putc('\r'); // ACK
#ifdef AVR109_EXTENSION
	    } else if (ch == 'r') { // Read Lock Bits
		putc(do_avrisp4(0x58,0x00,0x00,0x00));
	    } else if (ch == 'F') { // Read L-Fuses
		putc(do_avrisp4(0x50,0x00,0x00,0x00));
	    } else if (ch == 'N') { // Read H-Fuses
		putc(do_avrisp4(0x58,0x08,0x00,0x00));
	    } else if (ch == 'Q') { // Read EXT-Fuses
		putc(do_avrisp4(0x58,0x08,0x00,0x00));
	    } else if (ch == 'g') { // Read memory block
		size = getc() << 8;	// high byte of buffersize
		size |= getc();	// low byte
		ch = getc();   // memory type

		if (ch == 'E') { // EEPROM

		    while (size--) {
			putc(do_avrisp2(0xA0,0x00));
			addr.word++;
		    }

	        } else if (!(addr.lo & 1) && !(size & 1)) { // FLASH
		    addr.word >>= 1;
		    size >>= 1;
		    while (size--) {
			putc(do_avrisp2(0x20,0x00)); // Read low byte
			putc(do_avrisp2(0x28,0x00)); // Read high byte
			addr.word++;
		    }
		    addr.word <<= 1;
		} else {
		    while (size--) {
		   	 putc(0);
		    }
		}

	    } else if (ch == 'B') { // Programm with buffer

		size = getc() << 8;	// high byte of buffersize
		size |= getc();	// low byte
		ch = getc();   // memory type

		if (ch == 'E')	{ // EEPROM
		    while (size--) {
			do_avrisp2(0xC1,getc());
			if ((addr.lo & 0x3) == 0x3) {
			    do_avrisp2(0xC2,getc());
			    mdelay(twd_eeprom);
		 	}
			addr.word++;
		    }
		    putc('\r'); // ACK

		} else if (ch == 'F' && !(addr.lo & ((target_pagesize<<1) -1))
				&& !(size & 1) ) { // FLASH 
		    addr.word >>= 1;
		    size >>= 1;
		    while (size--) {
			do_avrisp2(0x40,getc()); // Write low byte
			do_avrisp2(0x48,getc()); // Write high byte
			if (size == 0) {
			    do_avrisp2(0x4C,0x00); // Issue Page Write
			    mdelay(twd_flash);
			}
		        addr.word++;
		    }
		    addr.word <<= 1;
		    putc('\r'); // ACK

		} else {
		    putc(0); // Unsupported memory type or wrong params
		}
#endif
	    } else {
		putc(0); // Unsupported memory type or wrong params
	    }
	} else if (extmode) {
	    uint8_t size,i;
	    uint8_t dev;
	    int8_t r;
	    r = -1; // device is not existed
	    dev = getc();
	    size = getc();
	    if (ch == 's' ) { // Send
		for (i=0; i<size && i< EXT_BUFSIZE; i++) {
		    ext_buf[i] = getc();
		}
		for (; i<size; i++) {
			getc();
		}
		size = i;
		if (dev == DEV_SYS) {
#ifdef ADC_EXTENSION
		} else if (dev == DEV_ADC) {
		    r = adc_send(size);
#endif
#ifdef OSC_EXTENSION
		} else if (dev == DEV_OSC) {
		    r = osc_send(size);
#endif
#ifdef FCNT_EXTENSION
		} else if (dev == DEV_FCNT) {
		    r = fcnt_send(size);
#endif
#ifndef NO_DEV_ECHO
	        } else if (dev == DEV_ECHO) {
		    r = size;
#endif
#ifdef I2C_EXTENSION
		} else if (DEV_I2C <= dev && dev <= 0x7f) {
		    r = i2c_send(dev, size);
#endif
		}
		putc(r);
	    } else if (ch == 'r') { // Recv
		 if (size > EXT_BUFSIZE)
			 size = EXT_BUFSIZE;
		if (dev == DEV_SYS) {
		    ext_buf[0] = PORTA;
		    r = 1;
#ifdef ADC_EXTENSION
	        } else if (dev == DEV_ADC) {
		    r = adc_recv(size);
#endif
#ifdef OSC_EXTENSION
	        } else if (dev == DEV_OSC) {
		    r = osc_recv(size);
#endif
#ifdef FCNT_EXTENSION
	        } else if (dev == DEV_FCNT) {
		    r = fcnt_recv(size);
#endif
#ifndef NO_DEV_ECHO
	        } else if (dev == DEV_ECHO) {
		    r = size;
#endif
#ifdef I2C_EXTENSION
		} else if (DEV_I2C <= dev && dev <= 0x7f) {
		    r = i2c_recv(dev, size);
#endif
		}
		putc(r);
		for (i=0; i<r; i++) {
		    putc(ext_buf[i]);
	        }
	   } else {
		putc(r);
	   }
	} else if (ch == 'e') { // Enter Extent mode 
		extmode = 1;
#ifdef I2C_EXTENSION
		init_i2c();
#endif
#ifdef ADC_EXTENSION
		init_adc();
#endif
#ifdef OSC_EXTENSION
		init_osc();
#endif

		putc('Y');
	} else {
		putc('?');
	}
    }
    return 0;
}

